from flask import request
from flask.signals import got_request_exception
try:
  from flask_login import current_user
except ImportError:
  flask_login_imported = False
else:
  flask_login_imported = True

from .notifier import Notifier


def init_app(app):
  if 'pybrake' in app.extensions:
    raise ValueError('pybrake is already injected')
  if 'PYBRAKE' not in app.config:
    raise ValueError("app.config['PYBRAKE'] is not defined")

  app.extensions['pybrake'] = Notifier(**app.config['PYBRAKE'])
  got_request_exception.connect(_handle_exception, sender=app)

  return app


def _handle_exception(sender, exception, **_):
  notifier = sender.extensions['pybrake']

  notice = notifier.build_notice(exception)
  ctx = notice['context']
  ctx['method'] = request.method
  ctx['url'] = request.url

  try:
    user_addr = request.access_route[0]
  except IndexError:
    user_addr = request.remote_addr
  if user_addr:
    ctx['userAddr'] = user_addr

  if flask_login_imported and current_user.is_authenticated:
    user = dict(
      id=current_user.get_id(),
    )
    for s in ['username', 'name']:
      if hasattr(current_user, s):
        user[s] = getattr(current_user, s)
    ctx['user'] = user

  notice['params']['request'] = dict(
    form=request.form,
    json=request.json,
    files=request.files,
    cookies=request.cookies,
    headers=dict(request.headers),
    environ=request.environ,
    blueprint=request.blueprint,
    url_rule=request.url_rule,
    view_args=request.view_args,
  )

  notifier.send_notice(notice)



from .notifier import Notifier
from .logging import LoggingHandler, _LOG_RECORD_ATTRS_TO_COPY
from .test_helper import build_logging_record_error, build_logging_record_exception


def test_logging_handler_record_error():
  record = build_logging_record_error('hello', extra=dict(foo='bar'))

  h = LoggingHandler(notifier=Notifier())
  notice = h.build_notice(record)

  errors = notice['errors']
  assert len(errors) == 1

  error = errors[0]
  assert error['type'] == 'test'
  assert error['message'] == 'hello'

  backtrace = error['backtrace']
  assert len(backtrace) == 1

  frame = backtrace[0]
  assert frame['file'] == '/PROJECT_ROOT/pybrake/test_helper.py'
  assert frame['function'] == 'build_logging_record_error'
  assert frame['line'] == 14
  assert frame['code'] == {
    12: 'def build_logging_record_error(*args, **kwargs):',
    13: '  logger, dh = logger_dummy_handler()',
    14: '  logger.error(*args, **kwargs)',
    15: '  return dh.record',
    16: '',
  }

  ctx = notice['context']
  assert ctx['severity'] == 'ERROR'
  assert ctx['component'] == 'test_helper'

  params = notice['params']
  assert params['extra'] == {'foo': 'bar'}
  for attr in _LOG_RECORD_ATTRS_TO_COPY:
    assert attr in params


def test_logging_handler_record_exception():
  record = build_logging_record_exception()

  h = LoggingHandler(notifier=Notifier())
  notice = h.build_notice(record)

  errors = notice['errors']
  assert len(errors) == 1

  error = errors[0]
  assert error['type'] == 'ValueError'
  assert error['message'] == 'hello'

  backtrace = error['backtrace']
  assert len(backtrace) == 1

  frame = backtrace[0]
  assert frame['file'] == '/PROJECT_ROOT/pybrake/test_helper.py'
  assert frame['function'] == 'build_logging_record_exception'
  assert frame['line'] == 21
  assert frame['code'] == {
    19: '  logger, dh = logger_dummy_handler()',
    20: '  try:',
    21: "    raise ValueError('hello')",
    22: '  except ValueError as err:',
    23: '    logger.exception(err)',
  }



from functools import lru_cache

from .utils import logger


try:
  from django.conf import settings
  _DJANGO = True
except ImportError:
  _DJANGO = False


@lru_cache()
def get_global_notifier():
  if not _DJANGO:
    return None

  if not hasattr(settings, 'AIRBRAKE'):
    if settings.DEBUG:
      logger.info('pybrake is not configured - set settings.AIRBRAKE')
      return None

  from .notifier import Notifier
  return Notifier(**settings.AIRBRAKE)



import os
from functools import lru_cache

from .utils import logger


@lru_cache(maxsize=1000)
def get_git_revision(dirpath):
  try:
    return _get_git_revision(dirpath)
  except (OSError, IOError) as err:
    logger.error('get_git_revision failed: %s', err)
    return None


def _get_git_revision(dirpath):
  head = get_git_head(dirpath)

  prefix = 'ref: '
  if not head.startswith(prefix):
    return head
  head = head[len(prefix):]

  ref_file = os.path.join(dirpath, '.git', head)
  try:
    with open(ref_file) as f:
      return f.read().rstrip()
  except (OSError, IOError):
    pass

  refs_file = os.path.join(dirpath, '.git', 'packed-refs')
  with open(refs_file) as f:
    for line in f:
      if not line or line[0] == '#' or line[0] == '^':
        continue

      parts = line.rstrip().split(' ')
      if parts[1] == head:
        return parts[0]

  return None


def get_git_head(dirpath):
  head_file = os.path.join(dirpath, '.git', 'HEAD')
  with open(head_file) as f:
    return f.read().rstrip()



import os
import sys
import platform
import socket
import urllib.request
import urllib.error
from concurrent import futures
import json
import time

from .notice import jsonify_notice
from .code_hunks import get_code_hunk
from .git import get_git_revision
from .utils import logger
from .version import version


_ERR_IP_RATE_LIMITED = 'IP is rate limited'

_AIRBRAKE_URL_FORMAT = '{}/api/v3/projects/{}/notices'

_CONTEXT = dict(
  notifier=dict(
    name='pybrake',
    version=version,
    url='https://github.com/airbrake/pybrake',
  ),
  os=platform.platform(),
  language='Python/%s' % platform.python_version(),
  hostname=socket.gethostname(),
  versions=dict(
    python=platform.python_version(),
  ),
)


class Notifier:
  def __init__(self,
               project_id=0,
               project_key='',
               host='https://api.airbrake.io',
               **kwargs):
    self._filters = []
    self._rate_limit_reset = 0
    self._max_queue_size = kwargs.get('max_queue_size', 1000)
    self._thread_pool = None

    self._airbrake_url = _AIRBRAKE_URL_FORMAT.format(host, project_id)
    self._airbrake_headers = {'authorization': 'Bearer ' + project_key,
                              'content-type': 'application/json'}

    self._context = _CONTEXT.copy()
    self._context['rootDirectory'] = kwargs.get('root_directory', os.getcwd())

    rev = get_git_revision(self._context['rootDirectory'])
    if rev is not None:
      self._context['revision'] = rev

    if 'environment' in kwargs:
      self._context['environment'] = kwargs['environment']

  def close(self):
    if self._thread_pool is not None:
      self._thread_pool.shutdown()

  def add_filter(self, filter_fn):
    """Appends filter to the list.

    Filter is a function that accepts notice. Filter can modify passed
    notice or return None if notice must be ignored.
    """
    self._filters.append(filter_fn)

  def notify_sync(self, err):
    """Notifies Airbrake about exception.

    Under the hood notify is a shortcut for build_notice and send_notice.
    """
    notice = self.build_notice(err)
    return self.send_notice_sync(notice)

  def build_notice(self, err):
    """Builds Airbrake notice from the exception."""
    notice = dict(
      errors=self._build_errors(err),
      context=self._build_context(),
      params=dict(
        sys_executable=sys.executable,
        sys_path=sys.path,
      ),
    )
    return notice

  def send_notice_sync(self, notice):
    """Sends notice to Airbrake.

    It returns notice with 2 possible new keys:
    - {'id' => str} - notice id on success.
    - {'error' => str|Exception} - error on failure.
    """
    for fn in self._filters:
      r = fn(notice)
      if r is None:
        notice['error'] = 'notice is filtered out'
        return notice
      notice = r

    if time.time() < self._rate_limit_reset:
      notice['error'] = _ERR_IP_RATE_LIMITED
      return notice

    data = jsonify_notice(notice)
    req = urllib.request.Request(self._airbrake_url,
                                 data=data,
                                 headers=self._airbrake_headers)

    try:
      resp = urllib.request.urlopen(req, timeout=5)
    except urllib.error.HTTPError as err:
      resp = err
    except Exception as err: # pylint: disable=broad-except
      notice['error'] = err
      logger.error(notice['error'])
      return notice

    try:
      body = resp.read()
    except IOError as err:
      notice['error'] = err
      logger.error(notice['error'])
      return notice

    if not (200 <= resp.code < 300 or 400 <= resp.code < 500):
      notice['error'] = 'unexpected Airbrake response status code'
      notice['error_info'] = dict(code=resp.code, body=body)
      logger.error(notice['error'])
      return notice

    if resp.code == 429:
      return self._rate_limited(notice, resp)

    try:
      body = body.decode('utf-8')
    except UnicodeDecodeError as err:
      notice['error'] = err
      logger.error(notice['error'])
      return notice

    try:
      data = json.loads(body)
    except ValueError as err: # json.JSONDecodeError requires Python 3.5+
      notice['error'] = err
      logger.error(notice['error'])
      return notice

    if 'id' in data:
      notice['id'] = data['id']
      return notice

    if 'message' in data:
      notice['error'] = data['message']
      logger.error(notice['error'])
      return notice

    notice['error'] = 'unexpected Airbrake response'
    notice['error_info'] = dict(data=data)
    logger.error(notice['error'])
    return notice

  def _rate_limited(self, notice, resp):
    v = resp.headers.get('X-RateLimit-Delay')
    if v is None:
      notice['error'] = 'X-RateLimit-Delay header is missing'
      return notice

    try:
      delay = int(v)
    except ValueError as err:
      notice['error'] = err
      return notice

    self._rate_limit_reset = time.time() + delay

    notice['error'] = _ERR_IP_RATE_LIMITED
    return notice

  def notify(self, err):
    """Asynchronously notifies Airbrake about exception from separate thread.

    Returns concurrent.futures.Future.
    """
    notice = self.build_notice(err)
    return self.send_notice(notice)

  def send_notice(self, notice):
    """Asynchronously sends notice to Airbrake from separate thread.

    Returns concurrent.futures.Future.
    """
    pool = self._get_thread_pool()
    if pool._work_queue.qsize() >= self._max_queue_size:
      notice['error'] = 'queue is full'
      f = futures.Future()
      f.set_result(notice)
      return f
    return pool.submit(self.send_notice_sync, notice)

  def _build_errors(self, err):
    if isinstance(err, str):
      backtrace = self._build_backtrace_frame(sys._getframe())
      return [{
        'message': err,
        'backtrace': backtrace,
      }]

    errors = []
    while err != None:
      errors.append(self._build_error(err))
      err = err.__cause__ or err.__context__
    return errors

  def _build_error(self, err):
    backtrace = self._build_backtrace_tb(err.__traceback__)
    error = dict(
      type=err.__class__.__name__,
      message=str(err),
      backtrace=backtrace,
    )
    return error

  def _build_backtrace_tb(self, tb):
    backtrace = []
    for frame, lineno in self._walk_tb(tb):
      f = self._build_frame(frame, lineno)
      if f:
        backtrace.insert(0, f)
    return backtrace

  def _walk_tb(self, tb):
    while tb is not None:
      yield tb.tb_frame, tb.tb_lineno
      tb = tb.tb_next

  def _build_backtrace_frame(self, f):
    backtrace = []
    for frame, lineno in self._walk_stack(f):
      f = self._build_frame(frame, lineno)
      if f:
        backtrace.append(f)
    return backtrace

  def _walk_stack(self, f):
    while f is not None:
      yield f, f.f_lineno
      f = f.f_back

  def _build_frame(self, f, line):
    if f.f_locals.get('__traceback_hide__'):
      return None

    filename = f.f_code.co_filename
    func = f.f_code.co_name

    if filename.endswith('pybrake/notifier.py'):
      return None

    loader = f.f_globals.get('__loader__')
    module_name = f.f_globals.get('__name__')
    return self._frame_with_code(filename, func, line,
                                 loader=loader,
                                 module_name=module_name)

  # pylint: disable=too-many-arguments
  def _frame_with_code(self, filename, func, line, loader=None, module_name=None):
    frame = dict(
      file=self._clean_filename(filename),
      function=func,
      line=line,
    )

    lines = get_code_hunk(filename, line, loader=loader, module_name=module_name)
    if lines is not None:
      frame['code'] = lines

    return frame

  def _clean_filename(self, s):
    if '/lib/python' in s and '/site-packages/' in s:
      needed = '/site-packages/'
      ind = s.find(needed)
      if ind > -1:
        s = '[SITE_PACKAGES]/' + s[ind+len(needed):]
        return s

    s = s.replace(self._context['rootDirectory'], '/PROJECT_ROOT')
    return s

  def _build_context(self):
    ctx = self._context.copy()

    versions = ctx['versions']
    for name, mod in sys.modules.items():
      if name.startswith('_'):
        continue
      if hasattr(mod, '__version__'):
        versions[name] = mod.__version__

    return ctx

  def _get_thread_pool(self):
    if self._thread_pool is None:
      max_workers = (os.cpu_count() or 1) * 5
      self._thread_pool = futures.ThreadPoolExecutor(max_workers=max_workers)
    return self._thread_pool



def get_exception():
  try:
    raise ValueError('hello')
  except ValueError as err:
    return err


def build_notice_from_str(notifier, s):
  return notifier.build_notice(s)


def build_logging_record_error(*args, **kwargs):
  logger, dh = logger_dummy_handler()
  logger.error(*args, **kwargs)
  return dh.record


def build_logging_record_exception():
  logger, dh = logger_dummy_handler()
  try:
    raise ValueError('hello')
  except ValueError as err:
    logger.exception(err)
    return dh.record


def logger_dummy_handler():
  import logging

  class DummyHandler(logging.Handler):
    def __init__(self):
      super().__init__()
      self.record = None

    def emit(self, record):
      self.record = record

  logger = logging.getLogger('test')
  dh = DummyHandler()
  logger.addHandler(dh)

  return logger, dh



import logging

from .notifier import Notifier
from .global_notifier import get_global_notifier


def _log_record_attrs():
  record = logging.LogRecord('', logging.ERROR, '', 0, '', None, None)
  attrs = set()
  for k in vars(record).keys():
    attrs.add(k)
  attrs.add('message') # set by Formatter
  return attrs

_LOG_RECORD_ATTRS = _log_record_attrs()


_LOG_RECORD_ATTRS_TO_COPY = [
  'created',
  'msecs',
  'process',
  'processName',
  'relativeCreated',
  'thread',
  'threadName',
]


class LoggingHandler(logging.Handler):
  def __init__(self, notifier=None, level=logging.ERROR, **kwargs):
    logging.Handler.__init__(self, level=level)
    if notifier is None:
      notifier = get_global_notifier() or Notifier(**kwargs)
    self._notifier = notifier

  def emit(self, record):
    try:
      notice = self.build_notice(record)
      self._notifier.send_notice(notice)
    except Exception: # pylint: disable=broad-except
      self.handleError(record)

  def build_notice(self, record):
    notice = dict(
      errors=[self._build_error(record)],
      context=self._build_context(record),
      params=self._build_params(record),
    )
    return notice

  def _build_error(self, record):
    if record.exc_info:
      return self._build_error_from_exc_info(record.exc_info)

    error = dict(
      type=record.name,
      message=record.msg,
      backtrace=self._build_backtrace(record),
    )
    return error

  def _build_error_from_exc_info(self, exc_info):
    cls, err, tb = exc_info
    backtrace = self._notifier._build_backtrace_tb(tb)
    error = dict(
      type=cls.__name__,
      message=str(err),
      backtrace=backtrace,
    )
    return error

  def _build_backtrace(self, record):
    frame = self._notifier._frame_with_code(
      record.pathname, record.funcName, record.lineno)
    backtrace = [frame]
    return backtrace

  def _build_context(self, record):
    ctx = dict(
      severity=record.levelname,
      component=record.module,
    )
    return ctx

  def _build_params(self, record):
    params = dict()

    extra = self._build_extra(record)
    if extra is not None:
      params['extra'] = extra

    for attr in _LOG_RECORD_ATTRS_TO_COPY:
      v = getattr(record, attr)
      if v is not None:
        params[attr] = v
    return params

  def _build_extra(self, record):
    extra = None
    for k, v in vars(record).items():
      if k not in _LOG_RECORD_ATTRS:
        if extra is None:
          extra = {k: v}
        else:
          extra[k] = v
    return extra



import functools

from django.conf import settings
from django.utils.module_loading import import_string

from .global_notifier import get_global_notifier


class AirbrakeMiddleware:
  def __init__(self, get_response):
    self.get_response = get_response
    self._notifier = get_global_notifier()

  def __call__(self, request):
    response = self.get_response(request)
    return response

  def process_exception(self, request, exception):
    if not self._notifier:
      return

    notice = self._notifier.build_notice(exception)

    ctx = notice['context']
    ctx['url'] = request.build_absolute_uri()

    user_addr = get_remote_addr(request)
    if user_addr:
      ctx['userAddr'] = user_addr

    req_filter = get_exception_reporter_filter(request)
    notice['params']['request'] = dict(
      scheme=request.scheme,
      method=request.method,
      GET=request.GET,
      POST=req_filter.get_post_parameters(request),
      META=dict(request.META),
      FILES=request.FILES,
      COOKIES=request.COOKIES,
      session=dict(request.session),
    )

    if request.user.is_authenticated:
      user = request.user
      user_info = dict(
        username=user.username,
        email=user.email,
      )

      names = [user.first_name, user.last_name]
      names = [x for x in names if x]
      if names:
        user_info['name'] = ' '.join(names)

      ctx['user'] = user_info

    self._notifier.send_notice(notice)


def get_remote_addr(request):
  x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
  if x_forwarded_for:
    return x_forwarded_for.split(',')[0]
  return request.META.get('REMOTE_ADDR')


@functools.lru_cache()
def get_default_exception_reporter_filter():
  # Instantiate the default filter for the first time and cache it.
  return import_string(settings.DEFAULT_EXCEPTION_REPORTER_FILTER)()


def get_exception_reporter_filter(request):
  default_filter = get_default_exception_reporter_filter()
  return getattr(request, 'exception_reporter_filter', default_filter)



import logging


def _get_logger():
  l = logging.getLogger('pybrake')
  fmter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

  sh = logging.StreamHandler()
  sh.setFormatter(fmter)
  l.addHandler(sh)

  return l

logger = _get_logger()



from functools import lru_cache


@lru_cache(maxsize=1000)
def get_code_hunk(filename, lineno, nlines=2, loader=None, module_name=None):
  lines = _get_lines_from_file(filename, loader=loader, module_name=module_name)
  if lines is None:
    return None

  start = max(0, lineno - 1 - nlines)
  end = min(lineno + nlines, len(lines))
  hunk = dict()
  for i, line in enumerate(lines[start:end], start + 1):
    if isinstance(line, bytes):
      line = line.decode('utf8', 'replace')
    hunk[i] = line.rstrip('\n')

  return hunk


@lru_cache(maxsize=10)
def _get_lines_from_file(filename, loader=None, module_name=None):
  if loader is not None and hasattr(loader, 'get_source'):
    try:
      source = loader.get_source(module_name)
    except ImportError:
      pass
    if source is not None:
      return source.splitlines()

  try:
    with open(filename, 'rb') as f:
      return f.read().splitlines()
  except (OSError, IOError):
    return None
